# coding=utf-8

from pytdx.parser.base import BaseParser
from pytdx.helper import get_datetime, get_volume, get_price
from collections import OrderedDict
import struct
import six

class GetCompanyInfoCategory(BaseParser):

    def setParams(self, market, code):
        if type(code) is six.text_type:
            code = code.encode("utf-8")

        pkg = bytearray.fromhex(u'0c 0f 10 9b 00 01 0e 00 0e 00 cf 02')
        pkg.extend(struct.pack(u"<H6sI", market, code, 0))
        self.send_pkg = pkg
    """

    10 00 d7 ee d0 c2 cc e1 ca be 00 00 ..... 36 30 30 33 30 30 2e 74 78 74 .... e8 e3 07 00 92 1f 00 00 .....

    10.... name
    36.... filename

    e8 e3 07 00 --- start
    92 1f 00 00 --- length

    """
    def parseResponse(self, body_buf):
        pos = 0
        (num, ) = struct.unpack("<H", body_buf[:2])
        pos += 2

        category = []



        def get_str(b):
            p = b.find(b'\x00')
            if p != -1:
                b = b[0: p]
            try:
                n = b.decode("gbk")
            except Exception as e:
                n = "unkown_str"
            return n

        for i in range(num):
            (name, filename, start, length) = struct.unpack(u"<64s80sII", body_buf[pos: pos+ 152])
            pos += 152
            entry = OrderedDict(
                [
                    ('name', get_str(name)),
                    ('filename', get_str(filename)),
                    ('start', start),
                    ('length', length),
                ]
            )
            category.append(entry)

        return category




#utf-8

import random
import threading
from functools import partial
from pytdx.log import DEBUG, log
import time
from collections import OrderedDict

"""
ips 应该还是一个 (ip ,port) 对的列表，如

[
    (ip1, port1),
    (ip2, port2),
    (ip3, port3),
]

"""

class BaseIPPool(object):

    def __init__(self, hq_class):
        self.hq_class = hq_class

    def setup(self):
        pass

    def teardown(self):
        pass

    def sync_get_top_n(self, num):
        pass

    def add_to_pool(self, ip):
        pass


class RandomIPPool(BaseIPPool):
    """
    获取一个随机的优先级列表
    """

    def __init__(self, hq_class, ips):
        """
        :param ips: ip should be a list
        """
        super(RandomIPPool, self).__init__(hq_class)
        self.ips = ips

    def get_ips(self):
        random.shuffle(self.ips)
        return self.ips

    def sync_get_top_n(self, num):
        ips= self.get_ips()
        return ips[:num]

    def add_to_pool(self, ip):
        if ip not in self.ips:
            self.ips.append(ip)


class AvailableIPPool(BaseIPPool):
    """
    测试可连接性，并根据连接速度排序
    我们启动一个新的线程，周期性的进行更新
    """

    def __init__(self, hq_class, ips):
        super(AvailableIPPool, self).__init__(hq_class)
        self.ips = ips
        self.sorted_ips = None
        self.worker_thread = None
        self.sorted_ips_lock = threading.Lock()
        self.stop_event = threading.Event()
        self.wait_interval = 20 * 60

    def setup(self):
        super(AvailableIPPool, self).setup()

        self.worker_thread = threading.Thread(target=self.run)
        self.worker_thread.start()

    def get_ips(self):
        if not self.sorted_ips:
            return self.ips
        else:
            return list(self.sorted_ips.values())

    def teardown(self):
        self.stop_event.set()
        if self.worker_thread.is_alive():
            self.worker_thread.join()
        self.worker_thread = None

    def run(self):
        log.debug("pool thread start ")
        while not self.stop_event.is_set():
            _available_ips = self.get_all_available_ips()
            sorted_keys = sorted(_available_ips)
            with self.sorted_ips_lock:
                self.sorted_ips = OrderedDict((key, _available_ips[key]) for key in sorted_keys)
            self.stop_event.wait(self.wait_interval)

    def get_all_available_ips(self):
        """
        循环测试所有连接的连接速度和有效性
        :return:
        """
        _available_ips = OrderedDict()
        for ip in self.ips:
            ip_addr, port = ip
            api = self.hq_class(multithread=False, heartbeat=False)
            try:
                with api.connect(ip_addr, port):
                    start_ts = time.time()
                    api.do_heartbeat()
                    end_ts = time.time()
                    diff_ts = end_ts - start_ts
                    _available_ips[diff_ts] = ip
                    log.debug("time diff is %f for %s" % (diff_ts, _available_ips))
            except Exception as e:
                log.debug("can not use %s:%d the exception is %s" % (ip_addr, port, str(e)))
                continue
        return _available_ips

    def sync_get_top_n(self, num):
        _ips = list(self.get_all_available_ips().values())
        return _ips[:min(len(_ips), num)]

    def add_to_pool(self, ip):
        if ip not in self.ips:
            self.ips.append(ip)


if __name__ == "__main__":
    from pytdx.hq import TdxHq_API
    from pytdx.config.hosts import hq_hosts
    import logging
    log.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    # create formatter
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    # add formatter to ch
    ch.setFormatter(formatter)
    log.addHandler(ch)

    ips = [(v[1], v[2]) for v in hq_hosts]
    pool = AvailableIPPool(TdxHq_API, ips)
    pool.wait_interval = 60 * 5
    pool.setup()
    sleep_time = 130
    log.debug("ready to sleep %d" % sleep_time )
    time.sleep(sleep_time)
    log.debug("sleep done")
    ips = pool.get_ips()
    log.debug(str(pool.get_ips()))
    log.debug("ready to teardown")
    pool.teardown()






# utf-8
from pytdx.log import DEBUG, log
from functools import partial
import time


## 调用单个接口，重试次数，超过次数则不再重试
DEFAULT_API_CALL_MAX_RETRY_TIMES = 20
## 重试间隔的休眠时间
DEFAULT_API_RETRY_INTERVAL = 0.2

class TdxHqApiCallMaxRetryTimesReachedException(Exception):
    pass

class TdxHqPool_API(object):
    """
    实现一个连接池的机制
    包含：

    1 1个正在进行数据通信的主连接
    2 1个备选连接，备选连接也连接到服务器，通过心跳包维持连接，当主连接通讯出现问题时，备选连接立刻转化为主连接, 原来的主连接返回ip池，并从ip池中选取新的备选连接
    3 m个ip构成的ip池，可以通过某个方法获取列表，列表可以进行排序，如果备选连接缺少的时候，我们根据排序的优先级顺序将其追加到备选连接
    """

    def __init__(self, hq_cls, ippool):
        self.hq_cls = hq_cls
        self.ippool = ippool
        """
        正在通信的客户端连接
        """
        self.api = hq_cls(multithread=True, heartbeat=True)
        """
        备选连接
        """
        self.hot_failover_api = hq_cls(multithread=True, heartbeat=True)

        self.api_call_max_retry_times = DEFAULT_API_CALL_MAX_RETRY_TIMES
        self.api_call_retry_times = 0
        self.api_retry_interval = DEFAULT_API_RETRY_INTERVAL


        # 对hq_cls 里面的get_系列函数进行反射
        log.debug("perform_reflect")
        self.perform_reflect(self.api)

    def perform_reflect(self, api_obj):
        # ref : https://stackoverflow.com/questions/34439/finding-what-methods-an-object-has
        method_names = [attr for attr in dir(api_obj) if callable(getattr(api_obj, attr))]
        for method_name in method_names:
            log.debug("testing attr %s" % method_name)
            if method_name[:3] == 'get' or method_name == "do_heartbeat" or method_name == 'to_df':
                log.debug("set refletion to method: %s", method_name)
                _do_hp_api_call = partial(self.do_hq_api_call, method_name)
                setattr(self, method_name, _do_hp_api_call)

    def do_hq_api_call(self, method_name, *args, **kwargs):
        """
        代理发送请求到实际的客户端
        :param method_name: 调用的方法名称
        :param args: 参数
        :param kwargs: kv参数
        :return: 调用结果
        """
        try:
            result = getattr(self.api, method_name)(*args, **kwargs)
            if result is None:
                log.info("api(%s) call return None" % (method_name,))
        except Exception as e:
            log.info("api(%s) call failed, Exception is %s" % (method_name, str(e)))
            result = None

        # 如果无法获取信息，则进行重试
        if result is None:
            if self.api_call_retry_times >= self.api_call_max_retry_times:
                log.info("(method_name=%s) max retry times(%d) reached" % (method_name, self.api_call_max_retry_times))
                raise TdxHqApiCallMaxRetryTimesReachedException("(method_name=%s) max retry times reached" % method_name)
            old_api_ip = self.api.ip
            new_api_ip = None
            if self.hot_failover_api:
                new_api_ip = self.hot_failover_api.ip
                log.info("api call from init client (ip=%s) err, perform rotate to (ip =%s)..." %(old_api_ip, new_api_ip))
                self.api.disconnect()
                self.api = self.hot_failover_api
            log.info("retry times is " + str(self.api_call_max_retry_times))
            # 从池里再次获取备用ip
            new_ips = self.ippool.get_ips()

            choise_ip = None
            for _test_ip in new_ips:
                if _test_ip[0] == old_api_ip or _test_ip[0] == new_api_ip:
                    continue
                choise_ip = _test_ip
                break

            if choise_ip:
                self.hot_failover_api = self.hq_cls(multithread=True, heartbeat=True)
                self.hot_failover_api.connect(*choise_ip)
            else:
                self.hot_failover_api = None
            # 阻塞0.2秒，然后递归调用自己
            time.sleep(self.api_retry_interval)
            self.do_hq_api_call(method_name, *args, **kwargs)
            self.api_call_retry_times += 1

        else:
            self.api_call_retry_times = 0

        return result

    def connect(self, ipandport, hot_failover_ipandport):
        log.debug("setup ip pool")
        self.ippool.setup()
        log.debug("connecting to primary api")
        self.api.connect(*ipandport)
        log.debug("connecting to hot backup api")
        self.hot_failover_api.connect(*hot_failover_ipandport)
        return self

    def disconnect(self):
        log.debug("primary api disconnected")
        self.api.disconnect()
        log.debug("hot backup api  disconnected")
        self.hot_failover_api.disconnect()
        log.debug("ip pool released")
        self.ippool.teardown()

    def close(self):
        """
        disconnect的别名，为了支持 with closing(obj): 语法
        :return:
        """
        self.disconnect()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


if __name__ == '__main__':

    from pytdx.hq import TdxHq_API
    from pytdx.pool.ippool import AvailableIPPool
    from pytdx.config.hosts import hq_hosts
    import random
    import logging
    import pprint
    log.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    # create formatter
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    # add formatter to ch
    ch.setFormatter(formatter)
    log.addHandler(ch)

    ips = [(v[1], v[2]) for v in hq_hosts]

    # 获取5个随机ip作为ip池
    random.shuffle(ips)
    ips5 = ips[:5]

    ippool = AvailableIPPool(TdxHq_API, ips5)

    primary_ip, hot_backup_ip = ippool.sync_get_top_n(2)

    print("make pool api")
    api = TdxHqPool_API(TdxHq_API, ippool)
    print("make pool api done")
    print("send api call to primary ip %s, %s" % (str(primary_ip), str(hot_backup_ip)))
    with api.connect(primary_ip, hot_backup_ip):
        ret = api.get_xdxr_info(0, '000001')
        print("send api call done")
        pprint.pprint(ret)



